#!/usr/bin/env python3
from sysrepo_api import sysrepo_api3 as api
import subprocess
import os
import platform

# system paths
TVE_REBOOT_DIR = '/config/_reboot/tve/now'

# Paths for modifications
GRUB_PATH = "/etc/default/grub"
UEFI_PATH = "/boot/efi/EFI/refind/refind.conf"
FSTAB_PATH = '/etc/fstab'
SYSTEM_CONF_PATH = '/etc/systemd/system.conf'
central_cbox_file="/config/central-box"


class SysrepoSystemResources(api.SysrepoAPI):
    _DoReboot = False
    rebootReason = ""

    def __init__(self,
                 app_name="system_resources",
                 logger=None,
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/system-resources:system-resources",
                 status_namespace="/ucpe:status/system-resources:system-resources",
                 run_as_subscriber=False,
                 priority=0,
                 upgrade=False):
        _DoReboot = False
        rebootReason = ""
        self._IsUpgrade = upgrade
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            logger=logger,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            status_namespace=status_namespace,
            run_as_subscriber=run_as_subscriber,
            priority=priority)


    def setBootReason(self, reason=""):
        self.logger.debug("{}".format(reason))
        SysrepoSystemResources.rebootReason = reason


    def setBootFlag(self, flag):
        self.logger.debug("setBootFlag set to {}".format(flag))
        SysrepoSystemResources._DoReboot = flag

    def getBootFlag(self):
        return SysrepoSystemResources._DoReboot

    def _request_tve_reboot(self):
        # Creation of TVE_REBOOT_DIR directory triggers the restart.
        # If the dir is present then remove it first, then create it. Otherwise the trigger will not sense the creation
        # Create dirs
        if self._IsUpgrade:
            self.logger.warning("Reboot requested. Not rebooting in upgrade sequence. Changes will get affect after reboot.")
            return
        try:
            self.logger.warning("%s" % SysrepoSystemResources.rebootReason)
            os.makedirs(TVE_REBOOT_DIR)
        except:
            self.logger.error(f"Error creating directory \"{TVE_REBOOT_DIR}\" for reboot. Directory already exists or cannot be created.")

    def _isin_xpath(self, xpath, keyword):
        if xpath.find(keyword) != -1:
            return True
        else:
            return False

    def _updateFromChangeSet(self, value, change_set, item):
        if change_set and item:
            for change in change_set:
                xpath = change['xpath']
                if xpath.find(item) > -1:
                    return change['new_value']
        return value

    def _isApplied(self, what, where, pref="", suf=""):
        try:
            # In case of wildcard *, get all files
            output = subprocess.check_output('find %s -type f' % where, shell=True).decode().strip()  # Get all present files
            files = output.strip().split('\n')  # Make a list with those files

            # Check how many files contains searched string
            okfiles = 0
            for _file in files:
                result = subprocess.check_output('grep -o "%s" %s' % (pref+what+suf, _file), shell=True).decode().strip()
                parameters = [parameter.strip() for parameter in result.split('\n')]
                if type(parameters) == str:
                    if what.strip() == parameters:
                        okfiles = okfiles + 1
                    else:
                        break
                else:
                    okparam = 0
                    for parameter in parameters:
                        if what.strip() == parameter.strip():
                            okparam = okparam + 1
                    if okparam > 0:
                        okfiles = okfiles + 1
                    else:
                        break

            # If all files are changed
            if okfiles == len(files):
                return True
        except:
            pass

        return False

    def _isUefi(self):
        retval = False
        try:
            # Get where is efibootmgr. Shell is needed to use PATH
            output = subprocess.check_output("which efibootmgr 2>/dev/null", shell=True).decode()
            if output != None:
                # Strip command only
                efibootmgr_cmd = output.rstrip()

                # Execute efibootmgr
                subprocess.check_call(efibootmgr_cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
                retval = True  # If we reach here that means that the uefi is present
        except:
            # If the execution of either "check_output" or "check_call" failed, then and exception will be thrown.
            # But retval is already set to value that shows that uefi is not present, so nothing to do more
            self.logger.debug("Fail getting UEFI, assuming GRUB")
            pass

        return retval

    def _getMachineCPUsCount(self):
        retval = -1
        try:
            with open("/sys/devices/system/cpu/present", 'r') as workingfile:
                # Get data as a long string
                cpuinfo_data = workingfile.read()
                retval = int(cpuinfo_data.strip().split("-")[1])
            workingfile.close()
        except:
            pass  # retval is already set to error value

        return retval

    def _getMachineMemory(self):
        retval = -1
        try:
            with open("/proc/meminfo", 'r') as workingfile:
                # Get data as a long string
                meminfo_data = workingfile.read()
                # Find the last processor
                memstr = "MemTotal:"
                meminfo_idx_begin = meminfo_data.find(memstr) + (len(memstr) + 1)
                # Find the end of the line
                meminfo_idx_end = meminfo_data.find("kB", meminfo_idx_begin)
                # Cut that part
                meminfo_data_chunk = meminfo_data[meminfo_idx_begin:meminfo_idx_end]
                # Cut only numbers after ":"
                memory = int(meminfo_data_chunk)

            workingfile.close()
            retval = memory / 1024  # In MB
        except:
            pass  # retval is already set to error value

        return retval

    def _getHostCPUsAndIsolCPUsLists(self, excludeHostCPUsList=None, excludeDPDKCPUsList=None):
        cpus_count = self._getMachineCPUsCount()
        if cpus_count > -1:
            cpuslist = range(0, cpus_count + 1)  # Create zero based list of CPUs

            HostCPUsList = []
            newCPUList = []
            newCPUList_DPDK = []
            # Generate list of CPU-s that excludes Host CPUs
            if excludeHostCPUsList != None and excludeHostCPUsList != "":
                # Guess separator and make list
                if excludeHostCPUsList.find(',') != -1:
                    HostCPUsList = excludeHostCPUsList.split(',')  # Do split based on comma separator
                else:
                    HostCPUsList = excludeHostCPUsList.split(' ')  # Do split based on space separator

                # Create a new list that does not contain elements from HostCPUs list
                newCPUList = [CPUs for CPUs in cpuslist if not str(CPUs) in HostCPUsList]

            # Exclude also DPDK CPUs
            if excludeDPDKCPUsList != None and excludeDPDKCPUsList != "" and excludeDPDKCPUsList != "none":
                # Guess separator and make list
                if excludeDPDKCPUsList.find(',') != -1:
                    DPDKCPUsList = excludeDPDKCPUsList.split(',')  # Do split based on comma separator
                else:
                    DPDKCPUsList = excludeDPDKCPUsList.split(' ')  # Do split based on space separator

                if len(newCPUList) > 0:
                    # If there are excluded Host CPUs
                    newCPUList_DPDK = [CPUs for CPUs in newCPUList if not str(CPUs) in DPDKCPUsList]
                else:
                    # If there are no excluded Host CPUs
                    newCPUList_DPDK = [CPUs for CPUs in cpuslist if not str(CPUs) in DPDKCPUsList]

            if len(HostCPUsList) > 0:
                if len(newCPUList_DPDK) > 0:
                    # return HostCPUs space separated list, IsolCPUs comma separated list, without host and/or dpdk cpus
                    return (' '.join(str(element) for element in HostCPUsList),
                            ','.join(str(element) for element in newCPUList_DPDK))
                elif len(newCPUList) > 0:
                    # return HostCPUs space separated list, IsolCPUs comma separated list, without host only cpus
                    return (' '.join(str(element) for element in HostCPUsList),
                            ','.join(str(element) for element in newCPUList))
            else:
                # return empty string for HostCPUs, IsolCPUs comma separated list which is machine CPUs
                return ('', ','.join(str(element) for element in cpuslist))

        # Return and empty strings for both HostCPUs and IsolCPUs
        return ('', '')

    def _GetHostCPUsCount(self, HostCPUsList):
        HostCPUsCount = -1
        if HostCPUsList != None and HostCPUsList != "":
            # Guess separator and make list
            if HostCPUsList.find(',') != -1:
                _HostCPUsList = HostCPUsList.split(',')  # Do split based on comma separator
            else:
                _HostCPUsList = HostCPUsList.split(' ')  # Do split based on space separator

            if len(_HostCPUsList) > 0:
                HostCPUsCount = len(_HostCPUsList)

        return HostCPUsCount

    def _fill_k8s_conf(self, HostCPUsAsLists, DPDKCPUsAsLists):
        res = False
        thefile="/var/lib/kubelet/config.yaml"

        if (os.path.isfile(central_cbox_file)):  #if we are central-cbox, don't touch CPU affinity
            #self.logger.debug("on Central Controller, ignore CPUAffinity changes!")
            return res

        if not os.path.isfile(thefile):
            return res

        src = "^reservedSystemCPUs: .*"
        dst = "reservedSystemCPUs: \\\"%s\\\"" % DPDKCPUsAsLists

        if os.system("egrep -q \"%s\" %s" %(src, thefile)) == 0:
            subprocess.call('sed -i "s/%s/%s/" %s' % (src, dst, thefile), shell=True)
        else:
            with open(thefile, 'a') as f:
                f.write("reservedSystemCPUs: \"%s\"\n" % DPDKCPUsAsLists)
        return res

    def _orgstyle_fill_systemd_system_conf(self, HostCPUsAsLists):
        res = False

        if (os.path.isfile(central_cbox_file)):  #if we are central-cbox, don't touch CPU affinity
            self.logger.debug("on Central Controller, ignore CPUAffinity changes!")
            return res

        src = ".*CPUAffinity=.*"
        dst = "CPUAffinity=%s" % HostCPUsAsLists

        if self._isApplied(dst, SYSTEM_CONF_PATH, ".*", ".*") == False:
            subprocess.call('sed -i "s/%s/%s/" %s' % (src, dst, SYSTEM_CONF_PATH), shell=True)
            if self._isApplied(dst, SYSTEM_CONF_PATH, ".*", ".*") == True:
                self.logger.debug("Success changing %s file" % SYSTEM_CONF_PATH)
                res = True
            else:
                self.logger.error("Error changing %s file" % SYSTEM_CONF_PATH)
                return False
        else:
            self.logger.debug("No need of changing %s file" % SYSTEM_CONF_PATH)
            self.logger.debug("CPUAffinity - %s" % HostCPUsAsLists)
            self.logger.debug("Searched string - %s" % dst)

        return res

    def Cpu_List_To_String( self, ListOfCPUs ):
        start=-10
        end=-10
        prev=-10
        counting=False
        output=[]
        if not ListOfCPUs or len(ListOfCPUs)==0:
            return ""

        ListOfCPUsInts = [int(i) for i in ListOfCPUs]
        ListOfCPUsSort = sorted(set(ListOfCPUsInts))
        for core in ListOfCPUsSort:
            core=int(core)
            if start==-10:
                prev=core
                start=core
                end=core
            elif core==(prev+1):
                if not counting:  # start counting
                    counting=True
                    start=prev

                end=core
                prev = core
            else: #no seq
                if counting: # seq finished
                    output.append("%s-%s" % (start,end))
                    start=core
                    prev=core
                    counting = False
                else:
                    output.append("%s" % prev)

                start=core
                prev=core

        if counting:
            output.append ("%s-%s" % (start,end))
        else:
            output.append("%s" % (prev))

        return ",".join(output)

    def _fill_rcunocbs_isolcpus(self, CPUsAsListsAsStr, isUefi):
        res = False

        if isUefi:
            sysres_path = UEFI_PATH
        else:
            sysres_path = GRUB_PATH

        if CPUsAsListsAsStr=="none":
            dst = "rcu_nocbs= "
        else:
            CPUsAsLists = CPUsAsListsAsStr.split(",")
            dst = "rcu_nocbs=%s" % self.Cpu_List_To_String(CPUsAsLists)


        src = "rcu_nocbs=[^[:space:]]\+"
        if self._isApplied(dst, sysres_path, "", " ") == False:
            subprocess.call('sed -i "s/%s/%s/" %s' % (src, dst, sysres_path), shell=True)
            if self._isApplied(dst, sysres_path, "", " ") == True:
                self.logger.debug("Success changing rcu_nocbs in %s file" % sysres_path)
                res = True
            else:
                self.logger.error("Error changing rcu_nocbs in %s file" % sysres_path)
                return False
        else:
            self.logger.debug("No need of changing rcu_nocbs in %s file" % sysres_path)
            self.logger.debug("IsolList - %s" % CPUsAsLists)
            self.logger.debug("Searched string - %s" % dst)

        src = "isolcpus=[^[:space:]]\+"

        if CPUsAsListsAsStr=="none":
            dst = "isolcpus= "
        else:
            dst = "isolcpus=%s" % self.Cpu_List_To_String(CPUsAsLists)

        if self._isApplied(dst, sysres_path, "", " ") == False:
            subprocess.call('sed -i "s/%s/%s/" %s' % (src, dst, sysres_path), shell=True)
            if self._isApplied(dst, sysres_path, "", " ") == True:
                self.logger.debug("Success changing isolcpus in %s file" % sysres_path)
                res = True
            else:
                self.logger.error("Error changing isolcpus in %s file" % sysres_path)
                return False
        else:
            self.logger.debug("No need of changing isolcpus in %s file" % sysres_path)
            self.logger.debug("IsolList - %s" % CPUsAsLists)
            self.logger.debug("Searched string - %s" % dst)

        return res

    def _orgstyle_fill_hpsize(self, HPSize, isUefi):
        res = False

        if (os.path.isfile(central_cbox_file)) and os.system('systemctl is-active --quiet kubelet.service')!=0:  #if we are central-cbox, update all endpoints with a hostname
            self.logger.debug("on Central Controller, ignore hpsize changes!")
            return res

        if isUefi:
            sysres_path = UEFI_PATH
        else:
            sysres_path = GRUB_PATH

        src = ".*hugetlbfs.*/c"
        dst = "hugetlbfs       /dev/hugepages  hugetlbfs       rw,pagesize=%dMB,mode=1770        0 0" % HPSize
        if self._isApplied(dst, FSTAB_PATH, ".*", ".*") == False:
            subprocess.call('sed -i "/%s\\%s" %s' % (src, dst, FSTAB_PATH), shell=True)
            if self._isApplied(dst, FSTAB_PATH, ".*", ".*") == True:
                self.logger.debug("Success changing %s file" % FSTAB_PATH)
                res = True
            else:
                self.logger.error("Error changing %s file" % FSTAB_PATH)
                return False
        else:
            self.logger.debug("No need of changing %s file" % FSTAB_PATH)
            self.logger.debug("HPSize - %d" % HPSize)
            self.logger.debug("Searched string - %s" % dst)

        src = "default_hugepagesz=[^[:space:]]\+ hugepagesz=[^[:space:]]\+"
        dst = "default_hugepagesz=%dM hugepagesz=%dM" % (HPSize, HPSize)
        if self._isApplied(dst, sysres_path) == False:
            subprocess.call('sed -i "s/%s/%s/" %s' % (src, dst, sysres_path), shell=True)
            if self._isApplied(dst, sysres_path) == True:
                self.logger.debug("Success changing [default_]hugepagesz in %s file" % sysres_path)
                res = True
            else:
                self.logger.error("Error changing [default_]hugepagesz in %s file" % sysres_path)
                return False
        else:
            self.logger.debug("No need of changing [default_]hugepagesz in %s file" % sysres_path)
            self.logger.debug("HPSize - %d" % HPSize)
            self.logger.debug("Searched string - %s" % dst)

        return res

    def _orgstyle_fill_hpcount(self, HPCount, isUefi):
        res = False
        if isUefi is True:
            sysres_path = UEFI_PATH
        else:
            sysres_path = GRUB_PATH

        if (os.path.isfile(central_cbox_file)) and os.system('systemctl is-active --quiet kubelet.service')!=0:  #if we are central-cbox, update all endpoints with a hostname
            self.logger.debug("on Central Controller, ignore hpcount changes!")
            return res

        src = " hugepages=\([0-9]\+\)"
        dst = " hugepages=%d" % HPCount
        if self._isApplied(dst, sysres_path) == False:
            subprocess.call('sed -i "s/%s/\%s/" %s' % (src, dst, sysres_path), shell=True)
            if self._isApplied(dst, sysres_path) == True:
                self.logger.debug("Success changing hugepages in %s file" % sysres_path)
                res = True
            else:
                self.logger.error("Error changing hugepages in %s file" % sysres_path)
                return False
        else:
            self.logger.debug("No need of changing of default_hugepagesz in %s file" % sysres_path)
            self.logger.debug("HPCount - %d" % HPCount)
            self.logger.debug("Searched string - %s" % dst)

        return res

    # An example of function that replaces the original style function (where sed is used)
    # Replacement functions with sed will be done on latter stage if needed
    def _fill_systemd_system_conf(self, CPUsAsLists):
        try:
            with open(SYSTEM_CONF_PATH, "r+") as file:
                # Get file
                filedata = file.readlines()

                # Search and replace
                isForChange = False
                for idx, filedata_line in enumerate(filedata):
                    if filedata_line.find("CPUAffinity=") != -1:
                        filedata[idx] = "CPUAffinity=%s\n" % CPUsAsLists
                        isForChange = True
                        break

                if isForChange:
                    # Go back in beginning
                    file.seek(0)
                    file.truncate()

                    # Write it
                    file.writelines(filedata)

            file.close()
        except:
            self.logger.error("Error opening %s file" % filename)

    def reboot_sequence_if_needed(self):
        if self.getBootFlag():
            self.setBootFlag(False)
            # There are GRUB related updates. Update GRUB
            if self._isUefi() is False:
                self.logger.info("Updating grub2!")
                subprocess.call('sudo update-grub2 > /dev/null 2>&1', shell=True)

            # Request reboot
            self.logger.warning("Requesting reboot. [%s]" % SysrepoSystemResources.rebootReason)
            self._request_tve_reboot()

    def _startupCase(self):
        isUefi = self._isUefi()
        HostCPUsList = self.get_host_reserved_cpus()
        DPDKCPUsList = None
        if HostCPUsList is not None or DPDKCPUsList is not None:
            self.logger.debug("HostCPUsList - {} -".format(HostCPUsList))
            self.logger.debug("DPDKCPUsList - {} -".format(DPDKCPUsList))
            CPUsLists = self._getHostCPUsAndIsolCPUsLists(HostCPUsList, DPDKCPUsList)
            AffinityCPUsList = CPUsLists[0]  # Host CPUs but space separated
            IsolCPUsList = CPUsLists[1]  # isolcpus CPUS but comma separated

            self.logger.debug("AffinityCPUsList - {} -".format(AffinityCPUsList))
            self.logger.debug("IsolCPUsList - {} -".format(IsolCPUsList))

            if len(HostCPUsList) > 0 and self._orgstyle_fill_systemd_system_conf(HostCPUsList) is True:
                self.setBootReason("Startup Changed HostCPUsList, requesting reboot")
                self.setBootFlag(True)

            if self._fill_rcunocbs_isolcpus(IsolCPUsList, isUefi) is True:
                self.setBootReason("Startup Changed DPDKCPUsList, requesting reboot")
                self.setBootFlag(True)

            self._fill_k8s_conf(HostCPUsList, IsolCPUsList)

        if self.is_dpdk() is True:
            DPDKCPUsList = self.get_dpdk_reserved_cpus()

        if HostCPUsList is not None or DPDKCPUsList is not None:
            CPUsLists = self._getHostCPUsAndIsolCPUsLists(HostCPUsList, DPDKCPUsList)
            IsolCPUsList = CPUsLists[1]  # isolcpus CPUS but comma separated
            # self.set_nova_vcpu_pin_set(IsolCPUsList)

        HPCount_key = self.get_huge_page_count()  # Get sysrepo data
        if HPCount_key is not None:
            HPCount = int(HPCount_key)
            if self._orgstyle_fill_hpcount(HPCount, isUefi) is True:
                self.setBootReason("Startup Changed HPCount, requesting reboot")
                self.setBootFlag(True)

        HPSize_key = self.get_huge_page_size()  # Get sysrepo data
        if HPSize_key is not None:
            HPSize = int(HPSize_key)
            if HPSize > 0:
                if self._orgstyle_fill_hpsize(HPSize, isUefi) is True:
                    self.setBootReason("Startup Changed HPSize, requesting reboot")
                    self.setBootFlag(True)

        isSubscription_key = self.get_host_oversubscribe_enabled()
        if isSubscription_key is not None:
            if isSubscription_key == 'true' or isSubscription_key is True:
                isSubscription = True
            elif isSubscription_key == 'false' or isSubscription_key is False:
                isSubscription = False

    def refresh_k8s_resources(self):
        HostCPUsList = self.get_host_reserved_cpus()
        DPDKCPUsList = None
        if HostCPUsList is not None or DPDKCPUsList is not None:
            CPUsLists = self._getHostCPUsAndIsolCPUsLists(HostCPUsList, DPDKCPUsList)
            AffinityCPUsList = CPUsLists[0]  # Host CPUs but space separated
            IsolCPUsList = CPUsLists[1]  # isolcpus CPUS but comma separated
            self._fill_k8s_conf(HostCPUsList, IsolCPUsList)


    def _xpathCase(self, xpath=None, change_set=None):
        isUefi = self._isUefi()

        item_host = "host-reserved-cpus"
        item_dpdk = "dpdk-reserved-cpus"
        if self._isin_xpath(xpath, item_host) or self._isin_xpath(xpath, item_dpdk):
            HostCPUsList = self.get_host_reserved_cpus()
            if self._isin_xpath(xpath, item_host):
                self.logger.debug("Xpath contains %s" % item_host)
                HostCPUsList = self._updateFromChangeSet(HostCPUsList, change_set, item_host)

            DPDKCPUsList = None
            if HostCPUsList is not None or DPDKCPUsList is not None or DPDKCPUsList is not "none":
                self.logger.debug("HostCPUsList %s" % HostCPUsList)
                self.logger.debug("DPDKCPUsList %s" % DPDKCPUsList)
                CPUsLists = self._getHostCPUsAndIsolCPUsLists(HostCPUsList, DPDKCPUsList)
                AffinityCPUsList = CPUsLists[0]  # Host CPUs but space separated
                IsolCPUsList = CPUsLists[1]  # isolcpus CPUS but comma separated

                self.logger.debug("AffinityCPUsList %s" % AffinityCPUsList)
                self.logger.debug("IsolCPUsList %s" % IsolCPUsList)

                # Update system files
                if self._orgstyle_fill_systemd_system_conf(AffinityCPUsList) is True:
                    self.setBootReason("Changed AffinityCPUsList, requesting reboot")
                    self.setBootFlag(True)

                if self._fill_rcunocbs_isolcpus(IsolCPUsList, isUefi) is True:
                    self.setBootReason("Changed IsolCPUsList, requesting reboot")
                    self.setBootFlag(True)

                self._fill_k8s_conf(HostCPUsList, IsolCPUsList)

            if self.is_dpdk() is True:
                DPDKCPUsList = self.get_dpdk_reserved_cpus()
                if self._isin_xpath(xpath, item_dpdk):
                    self.logger.debug("Xpath contains %s" % item_dpdk)
                    DPDKCPUsList = self._updateFromChangeSet(DPDKCPUsList, change_set, item_dpdk)

            if HostCPUsList is not None or DPDKCPUsList is not None:
                CPUsLists = self._getHostCPUsAndIsolCPUsLists(HostCPUsList, DPDKCPUsList)
                IsolCPUsList = CPUsLists[1]  # isolcpus CPUS but comma separated
                # self.set_nova_vcpu_pin_set(IsolCPUsList)

        item = "huge-page-count"
        if self._isin_xpath(xpath, item):
            HPCount_key = self.get_huge_page_count()  # Get sysrepo data
            HPCount_key = self._updateFromChangeSet(HPCount_key, change_set, item)
            if HPCount_key is not None:
                HPCount = int(HPCount_key)
                if self._orgstyle_fill_hpcount(HPCount, isUefi) is True:
                    self.setBootReason("Changed HPCount, requesting reboot")
                    self.setBootFlag(True)

        item = "huge-page-size"
        if self._isin_xpath(xpath, item):
            HPSize_key = self.get_huge_page_size()  # Get sysrepo data
            HPSize_key = self._updateFromChangeSet(HPSize_key, change_set, item)
            if HPSize_key is not None:
                HPSize = int(HPSize_key)
                if HPSize > 0:
                    if self._orgstyle_fill_hpsize(HPSize, isUefi) is True:
                        self.setBootReason("Changed HPSize, requesting reboot")
                        self.setBootFlag(True)

        if self._isin_xpath(xpath, "dpdk-reserved-cpus") or self._isin_xpath(xpath, "host-reserved-memory"):
            self.setBootReason("Changed dpdk-reserved-memory or host-reserved-memory, requesting reboot")
            self.setBootFlag(True)

        item = "host-oversubscribe-enabled"
        if self._isin_xpath(xpath, item):
            isSubscription_key = self.get_host_oversubscribe_enabled()
            isSubscription_key = self._updateFromChangeSet(isSubscription_key, change_set, item)
            if isSubscription_key is not None:
                if isSubscription_key == 'true' or isSubscription_key is True:
                    isSubscription = True
                elif isSubscription_key == 'false' or isSubscription_key is False:
                    isSubscription = False

    def apply_config(self, change_set=None, apply_startup=False):
        self.logger.debug("apply_config - BEGIN")
        self.logger.debug("apply_config - change_set {}".format(change_set))
        if apply_startup is True:
            self.logger.info("apply_config (startup)!")
            self._startupCase()
            self.reboot_sequence_if_needed()
        else:
            self.logger.info("apply_config (xpath)!")
            # We got some lists. So searching what is done.
            for idx in range(0, len(change_set)):
                if not change_set[idx]["xpath"].startswith(self.namespace):
                    continue

                # searching for MODIFIED or CREATED flag
                if change_set[idx]["operation"] == 'MODIFIED' or \
                   change_set[idx]["operation"] == 'CREATED':
                    # Triggered by parameter change
                    self._xpathCase(change_set[idx]["xpath"], change_set)

        self.logger.debug("apply_config - END")
        return True

    def post_apply_config(self, change_set):
        self.logger.debug("post_apply_config - BEGIN")
        self.reboot_sequence_if_needed()
        self.logger.debug("post_apply_config - END")

    def _get_kernel(self):
        kernel_name = ""
        try:
            kernel_name = platform.release()
        except:
            self.logger.error("Cannot get kernel name.")
            kernel_name = ""

        return kernel_name

    def _get_available_cpus(self):
        available_cpus = 0
        cpus_count = self._getMachineCPUsCount()
        if cpus_count > -1:
            available_cpus = cpus_count + 1 # Zero based cpu count

        return available_cpus

    def _get_available_memory(self):
        available_memory = 0
        memory_size = self._getMachineMemory()
        if memory_size > -1:
            available_memory = memory_size

        return available_memory

    def status_get_cb(self, xpath, private_data):
        data = { "ucpe:status": { "system-resources:system-resources": { "host-available-cpus": "{}".format(self._get_available_cpus()),
                                                                         "host-available-memory": "{}".format(int(self._get_available_memory())),
                                                                         "kernel": "{}".format(self._get_kernel())
                                                                       }
                                }
               }
        return data

    def set_huge_page_size(self, value=0):
        self.set_item("{}/huge-page-size".format(self.namespace), value)

    def get_huge_page_size(self):
        return self.get_item("{}/huge-page-size".format(self.namespace))

    def set_dpdk_reserved_cpus(self, value='2'):
        self.set_item("{}/dpdk-reserved-cpus".format(self.namespace), value)

    def get_dpdk_reserved_cpus(self):
        return self.get_item("{}/dpdk-reserved-cpus".format(self.namespace))

    def set_host_oversubscribe_enabled(self, value=False):
        self.set_item("{}/host-oversubscribe-enabled".format(self.namespace), value)

    def get_host_oversubscribe_enabled(self):
        return self.get_item("{}/host-oversubscribe-enabled".format(self.namespace))

    def set_huge_page_count(self, value=0):
        self.set_item("{}/huge-page-count".format(self.namespace), value)

    def get_huge_page_count(self):
        return self.get_item("{}/huge-page-count".format(self.namespace))

    def set_host_reserved_memory(self, value=1):
        self.set_item("{}/host-reserved-memory".format(self.namespace), value)

    def get_host_reserved_memory(self):
        return self.get_item("{}/host-reserved-memory".format(self.namespace))

    def set_host_reserved_cpus(self, value=0):
        self.set_item("{}/host-reserved-cpus".format(self.namespace), value)

    def get_host_reserved_cpus(self):
        return self.get_item("{}/host-reserved-cpus".format(self.namespace))

    def is_dpdk(self):
        return True # remove that when subscriber is subscribed to get changes to interfaces if it can be
    # self.namespace="/ucpe:config/interfaces:interface"
    # interfaces = self.get_running_config(full_path=False)
    # self.namespace="/ucpe:config/system-resources:system-resources"
    # for interface in interfaces:
    #     if 'dpdk' in interface.get('mode', ''):
    #         return True
    # return False
